#ifndef FOO_VIS_SIMPLE_SPECTRUM_GRADIENT_H
#define FOO_VIS_SIMPLE_SPECTRUM_GRADIENT_H

struct rgb_color_t
{
	double r, g, b;
};

struct hsv_color_t
{
	double h, s, v;
};

// r,g,b values are from 0 to 1
// h = [0,360], s = [0,1], v = [0,1]
//		if s == 0, then h = 0 (undefined)
static void convert_rgb_to_hsv(const rgb_color_t & in, hsv_color_t & out)
{

	double min, max, delta;

	min = in.r;
	if (min > in.g) min = in.g;
	if (min > in.b) min = in.b;
	max = in.r;
	if (max < in.g) max = in.g;
	if (max < in.b) max = in.b;
	out.v = max;				// v

	delta = max - min;

	if( max != 0 )
		out.s = delta / max;		// s
	else {
		// r = g = b = 0		// s = 0, v is undefined
		out.s = 0;
		out.h = 0;
		return;
	}

	if( in.r == max )
		out.h = ( in.g - in.b ) / delta;		// between yellow & magenta
	else if( in.g == max )
		out.h = 2 + ( in.b - in.r ) / delta;	// between cyan & yellow
	else
		out.h = 4 + ( in.r - in.g ) / delta;	// between magenta & cyan

	out.h *= 60;				// degrees
	if( out.h < 0 )
		out.h += 360;
}

static void convert_hsv_to_rgb(const hsv_color_t & in, rgb_color_t & out)
{
	int i;
	double f, p, q, t;
	double h;

	if( in.s == 0 ) {
		// achromatic (grey)
		out.r = out.g = out.b = in.v;
		return;
	}

	h = in.h / 60;			// sector 0 to 5
	i = (int)floor( h );
	f = h - i;			// factorial part of h
	p = in.v * ( 1 - in.s );
	q = in.v * ( 1 - in.s * f );
	t = in.v * ( 1 - in.s * ( 1 - f ) );

	switch( i ) {
	case 0:
		out.r = in.v;
		out.g = t;
		out.b = p;
		break;
	case 1:
		out.r = q;
		out.g = in.v;
		out.b = p;
		break;
	case 2:
		out.r = p;
		out.g = in.v;
		out.b = t;
		break;
	case 3:
		out.r = p;
		out.g = q;
		out.b = in.v;
		break;
	case 4:
		out.r = t;
		out.g = p;
		out.b = in.v;
		break;
	default:		// case 5:
		out.r = in.v;
		out.g = p;
		out.b = q;
		break;
	}
}

#ifdef DEBUG_COLOR_SPACE
static bool check_limits(double x, double lo, double hi)
{
	return (x >= lo) && (x <= hi);
}

static bool check_limits(const rgb_color_t & color)
{
	return check_limits(color.r,0.0,1.0) && check_limits(color.g,0.0,1.0) && check_limits(color.g,0.0,1.0);
}

static void assert_limits(const rgb_color_t & color, const char * name = 0)
{
	if (!check_limits(color))
	{
		string8 msg;
		msg = "values out of bound";
		if (name)
		{
			msg += " (";
			msg += name;
			msg += ")";
		}
		msg += ": RGB(";
		msg.add_float(color.r,6);
		msg += ", ";
		msg.add_float(color.g,6);
		msg += ", ";
		msg.add_float(color.b,6);
		msg += ")";
		console::warning(msg);
	}
}

static bool check_limits(const hsv_color_t & color)
{
	return check_limits(color.h,0,360) && check_limits(color.s,0,1) && check_limits(color.v,0,1);
}

static void assert_limits(const hsv_color_t & color, const char * name = 0)
{
	if (!check_limits(color))
	{
		string8 msg;
		msg = "values out of bound";
		if (name)
		{
			msg += " (";
			msg += name;
			msg += ")";
		}
		msg += ": HSV(";
		msg.add_float(color.h,6);
		msg += ", ";
		msg.add_float(color.s,6);
		msg += ", ";
		msg.add_float(color.v,6);
		msg += ")";
		console::warning(msg);
	}
}
#endif // DEBUG_COLOR_SPACE

class NOVTABLE gradient : public service_base
{
public:
	struct entry_t
	{
		double x;
		rgb_color_t color;
	};

	virtual void add_point(double x, const rgb_color_t & color) = 0;
	inline void add_point(double x, COLORREF c)
	{
		rgb_color_t d;
		d.r = (c & 0xFF) / 255.0;
		d.g = ((c >> 8) & 0xFF) / 255.0;
		d.b = ((c >> 16) & 0xFF) / 255.0;
		add_point(x, d);
	}

	virtual void get_color(double x, rgb_color_t & out) = 0;
	inline COLORREF get_color(double x)
	{
		rgb_color_t color;
		get_color(x, color);
		unsigned r, g, b;
		r = (unsigned)(color.r * 255);
		g = (unsigned)(color.g * 255);
		b = (unsigned)(color.b * 255);
		return RGB(r,g,b);
	}

	virtual const char * get_name() = 0;

	static const GUID & get_class_guid()
	{
		// {5035C780-BD64-11d7-A8B2-00E07D7A2153}
		static const GUID guid = 
		{ 0x5035c780, 0xbd64, 0x11d7, { 0xa8, 0xb2, 0x0, 0xe0, 0x7d, 0x7a, 0x21, 0x53 } };
		return guid;
	}

	static gradient * get(const char * name)
	{
		service_enum_t<gradient> e;
		gradient * ptr;
		for (ptr=e.first();ptr;ptr=e.next())
		{
			if (!strcmp(name, ptr->get_name()))
				return ptr;
			ptr->service_release();
		}
		return 0;
	}

};

template<class T>
class gradient_factory : public service_factory_t<gradient,T> {};

class NOVTABLE gradient_base : public gradient
{
protected:
	mem_block_list_t<entry_t> data;

	unsigned find_point(double x)
	{
		unsigned n, m = data.get_count();
		for (n=0;n<m;n++)
		{
			if (x <= data[n].x) break;
		}
		return n;
	}

	// 0 < f < 1
	virtual void blend(const rgb_color_t & in0, const rgb_color_t & in1, double f, rgb_color_t & out) = 0;

public:
	virtual void add_point(double x, const rgb_color_t & c)
	{
		entry_t entry;
		entry.x = x;
		entry.color = c;
		unsigned n = find_point(x);
		if (n == data.get_count())
		{
			data.add_item(entry);
		}
		else if (x == data[n].x)
		{
			data.replace_item(n, entry);
		}
		else
		{
			data.insert_item(entry, n);
		}
	}

	virtual void get_color(double x, rgb_color_t & out)
	{
		unsigned n;
		n = find_point(x);
		out.r = out.g = out.b = 0.0;
		if (n < data.get_count())
		{
			if (x == data[n].x)
				out = data[n].color;
			else if (n > 0)
			{
				double f = (x-data[n-1].x) / (data[n].x - data[n-1].x);
				blend(data[n-1].color,data[n].color,f,out);
			}
		}
	}
};

class gradient_rgb_linear : public gradient_base
{
protected:
	virtual void blend(const rgb_color_t & in0, const rgb_color_t & in1, double f, rgb_color_t & out)
	{
		out.r = in0.r*(1.0-f) + in1.r*f;
		out.g = in0.g*(1.0-f) + in1.g*f;
		out.b = in0.b*(1.0-f) + in1.b*f;
	}

public:
	virtual const char * get_name()
	{
		return "RGB linear";
	}
};

class NOVTABLE gradient_hsv : public gradient_base
{
private:
	virtual void blend(const rgb_color_t & in0, const rgb_color_t & in1, double f, rgb_color_t & out)
	{
		hsv_color_t hsv_in0, hsv_in1, hsv_out;

		convert_rgb_to_hsv(in0, hsv_in0);
		convert_rgb_to_hsv(in1, hsv_in1);

		blend(hsv_in0, hsv_in1, f, hsv_out);

		convert_hsv_to_rgb(hsv_out, out);
	}

protected:
	virtual void blend(const hsv_color_t & in0, const hsv_color_t & in1, double f, hsv_color_t & out) = 0;
};

class gradient_hsv_ccw : public gradient_hsv
{
protected:
	virtual void blend(const hsv_color_t & in0, const hsv_color_t & in1, double f, hsv_color_t & out)
	{
		if (in0.h < in1.h)
			out.h = in0.h*(1.0-f) + in1.h*f;
		else
		{
			out.h = (in0.h-360)*(1.0-f) + in1.h*f;
			if (out.h < 0) out.h += 360;
		}
		out.s = in0.s*(1.0-f) + in1.s*f;
		out.v = in0.v*(1.0-f) + in1.v*f;
	}

public:
	virtual const char * get_name()
	{
		return "HSV counterclockwise";
	}
};

class gradient_hsv_cw : public gradient_hsv
{
protected:
	virtual void blend(const hsv_color_t & in0, const hsv_color_t & in1, double f, hsv_color_t & out)
	{
		if (in0.h > in1.h)
			out.h = in0.h*(1.0-f) + in1.h*f;
		else
		{
			out.h = in0.h*(1.0-f) + (in1.h-360)*f;
			if (out.h < 0) out.h += 360;
		}
		out.s = in0.s*(1.0-f) + in1.s*f;
		out.v = in0.v*(1.0-f) + in1.v*f;
	}

public:
	virtual const char * get_name()
	{
		return "HSV clockwise";
	}
};

#endif // !FOO_VIS_SIMPLE_SPECTRUM_GRADIENT_H
