#define _WIN32_WINNT 0x0500
#include <windows.h>
#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"
#include "resource.h"
#include <commctrl.h>
#include "gradient.h"
#include "config.h"
#include "fft.h"

#include <math.h>

DECLARE_COMPONENT_VERSION("Simple spectrum", "0.1.6", 0);

typedef BOOL WINAPI SetLayeredWindowAttributesProc(HWND hwnd, COLORREF crKey, BYTE bAlpha, DWORD dwFlags);

static SetLayeredWindowAttributesProc * g_SetLayeredWindowAttributes;

static void init_layered_window() {
	g_SetLayeredWindowAttributes = 0;
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof (osvi);
	GetVersionEx(&osvi);
	if (osvi.dwMajorVersion >= 5) {
		HINSTANCE lib_user32 = uLoadLibrary("user32.dll");
		if (lib_user32) {
			g_SetLayeredWindowAttributes = 
				(SetLayeredWindowAttributesProc *)GetProcAddress(lib_user32, "SetLayeredWindowAttributes");
		}
	}
}

// number of interpolation steps for gradient
enum {
	VIS_COLOR_TABLE_SIZE	= 1024,
	VIS_COLOR_TABLE_MAX		= (VIS_COLOR_TABLE_SIZE-1),
};

cfg_int cfg_interpolation("interpolation", 0);
cfg_int cfg_toggle_show_caption("show caption", 1);
cfg_int cfg_toggle_use_alpha("use alpha", 0);
cfg_int cfg_alpha("alpha", 0);

class canvas;

// our window
static HWND g_wnd;
// size of out window
static RECT g_rect;
// precomputed gradient
static COLORREF g_colors[VIS_COLOR_TABLE_SIZE];
// next column to be drawn
static unsigned g_column;

// setting to enable this vis
static cfg_int cfg_enabled("enabled",0);

// for remembering the window position
static cfg_struct_t<WINDOWPLACEMENT> cfg_window_placement("window_placement",0);

// for detection of changes in settings
static unsigned g_last_nch = 0, g_last_spectrum_size = 0, g_last_srate = 0;
static int g_last_range = -1;
// precomputed frequency scaling, contains y-values, see recalc_grid function
static mem_block_t<int> g_grid;

// for closing the window 2 seconds after flushing
static unsigned g_timer = 0;

class canvas
{
	HBITMAP bmp;
	int width, height;
public:
	canvas() : bmp(0), width(-1), height(-1) {resize(0,0);}
	~canvas() {	if (bmp) {DeleteObject(bmp); bmp=0;}	}
	
	bool resize(int w, int h)
	{
		if (w != width || h != height)
		{
			HBITMAP bmp_new = 0;
			struct
			{
				BITMAPINFOHEADER bmi;
				DWORD colors[2];//edit if you want more/less clors (24bit doesnt need colortable)
			} bi;
			memset(&bi,0,sizeof(bi));
			bi.bmi.biSize = sizeof(bi.bmi);
			bi.bmi.biWidth = w;
			bi.bmi.biHeight = h;
			bi.bmi.biPlanes = 1;
			bi.bmi.biBitCount = 32;//edit if you want more colors
			bi.bmi.biCompression = BI_RGB;
			bi.colors[0]=0;//edit if you want different colors
			bi.colors[1]=0xFFFFFF;

			void * p_bits;
			bmp_new = CreateDIBSection(0,(BITMAPINFO*)&bi,DIB_RGB_COLORS,&p_bits,0,0);
			if (!bmp_new) return false;

			HBRUSH background;
			background = (HBRUSH)GetStockObject(BLACK_BRUSH);
			HDC dc = CreateCompatibleDC(0);
			HGDIOBJ meh = SelectObject(dc,bmp_new);
			FillRect(dc,&g_rect,background);

			if (bmp)
			{
#if 0
				HDC dc2 = CreateCompatibleDC(dc);
				HGDIOBJ meh2 = SelectObject(dc2,bmp);
				int width_min = (w<width) ? w : width;
				StretchBlt(dc,0,0,width_min,h,dc2,0,0,width_min,height,SRCCOPY);
				SelectObject(dc2, meh2);
				DeleteDC(dc2);
#endif
				DeleteObject(bmp);
			}

			SelectObject(dc,meh);
			DeleteDC(dc);
			bmp = bmp_new;
			width = w;
			height = h;
			return true;
		}
		return false;
	}

	operator HBITMAP() {return bmp;}
};

// our backbuffer
static canvas g_display;

class display_mode_notify : public cfg_var_notify
{
	virtual void on_var_change(const char * name, cfg_var * var)
	{
		if (var == &cfg_display_mode)
		{
			if (g_wnd)
			{
				switch (cfg_display_mode)
				{
				case DISPLAY_ALL_CHANNELS:
					uSetWindowText(g_wnd, "Simple spectrum (all channels)");
					break;
				case DISPLAY_MIXDOWN:
					uSetWindowText(g_wnd, "Simple spectrum (mixdown)");
					break;
				case DISPLAY_DIFFERENCE:
					uSetWindowText(g_wnd, "Simple spectrum (mean squared difference)");
					break;
				}
			}
		}
	}
public:
	display_mode_notify(cfg_var * var) : cfg_var_notify(var) {}
};

static display_mode_notify g_display_mode_notify(&cfg_display_mode);

static mem_block_t<double> g_fine_grid;

// channels are stored in g_grid from bottom to top, as the calculation is easier this way
static void recalc_grid(unsigned nch, unsigned spectrum_size)
{
	if (nch == g_last_nch && spectrum_size == g_last_spectrum_size && cfg_range == g_last_range)
		return;
	if (cfg_range == 0)
	{
		unsigned n, m = nch*spectrum_size;
		g_grid.set_size(m+1); // indices 0..m
		for (n=0;n<m;n++)
		{
			g_grid[n] = g_rect.bottom-MulDiv(g_rect.bottom,n,m);
		}
	}
	else
	{
		double x = (double)cfg_range/100.0;
		double range = exp(x)-1;
		double f = 1.0 / (double)nch / x;
		unsigned ch;
		g_grid.set_size(nch*spectrum_size+1);
		for (ch=0;ch<nch;ch++)
		{
			unsigned n;
			for (n=0;n<spectrum_size;n++)
			{
				g_grid[ch*spectrum_size+n] = (int)(g_rect.bottom * (1 - (double)ch/(double)nch - log((double)n/(double)spectrum_size*range+1)*f));
			}
		}
	}
	g_grid[nch*spectrum_size] = 0;
	g_last_nch = nch;
	g_last_spectrum_size = spectrum_size;
	g_last_range = cfg_range;
}

// channels are stored in g_fine_grid from bottom to top, as the calculation is easier this way
static void recalc_fine_grid(unsigned nch, unsigned spectrum_size)
{
	if (nch == g_last_nch && spectrum_size == g_last_spectrum_size && cfg_range == g_last_range)
		return;
	if (cfg_range == 0)
	{
		unsigned n, m = nch*spectrum_size;
		g_fine_grid.set_size(m+1); // indices 0..m
		for (n=0;n<m;n++)
		{
			g_fine_grid[n] = g_rect.bottom * (1.0 - (double)n / (double)m);
		}
	}
	else
	{
		double x = (double)cfg_range/100.0;
		double range = exp(x)-1;
		double f = 1.0 / (double)nch / x;
		unsigned ch;
		g_fine_grid.set_size(nch*spectrum_size+1);
		for (ch=0;ch<nch;ch++)
		{
			unsigned n;
			for (n=0;n<spectrum_size;n++)
			{
				g_fine_grid[ch*spectrum_size+n] = g_rect.bottom * (1.0 - (double)ch/(double)nch - log((double)n/(double)spectrum_size*range+1)*f);
			}
		}
	}
	g_fine_grid[nch*spectrum_size] = 0.0;
	g_last_nch = nch;
	g_last_spectrum_size = spectrum_size;
	g_last_range = cfg_range;
}

void g_close()
{
	if (g_wnd)
	{
		DestroyWindow(g_wnd);
		g_wnd=0;
	}
	if (g_timer)
	{
		KillTimer(0,g_timer);
		g_timer=0;
	}
}

static void CALLBACK TimerProc(HWND wnd,UINT msg,UINT id,DWORD time)
{
	g_close();
}

void g_flush()
{
	if (g_timer)
	{
		KillTimer(0,g_timer);
		g_timer = 0;
	}
	if (g_wnd)
	{
		g_timer = SetTimer(0,666,2000,TimerProc);
	}
}

static LRESULT WINAPI g_windowproc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);

static void create_vis_window() {
	static bool g_inited;
	static const char g_class_name[] = "foo_vis_simple_spectrum.dll window class";
	if (!g_inited)
	{
		g_inited = true;
		uWNDCLASS wc;
		memset(&wc,0,sizeof(wc));
		wc.style = CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS;
		wc.lpfnWndProc = g_windowproc;
		wc.hInstance = core_api::get_my_instance();
		wc.hCursor = uLoadCursor(0,IDC_ARROW);
		wc.hbrBackground = 0;
		wc.lpszClassName = g_class_name;
		uRegisterClass(&wc);
	}
	uCreateWindowEx(WS_EX_TOOLWINDOW | ((cfg_toggle_use_alpha && g_SetLayeredWindowAttributes) ? WS_EX_LAYERED : 0),
		g_class_name,"Test visualization 2",
		WS_POPUP|WS_THICKFRAME| (cfg_toggle_show_caption ? WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU : 0),
		CW_USEDEFAULT,CW_USEDEFAULT,
		512,512,
		core_api::get_main_window(),0,core_api::get_my_instance(),0);
}

static LRESULT WINAPI g_windowproc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{
	static bool dragging = false;
	static POINT drag_start;
	switch(msg)
	{
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			BeginPaint(wnd, &ps);
			HDC dc = ps.hdc;

			// spectrum
			HDC dc_bmp = CreateCompatibleDC(0);
			HGDIOBJ old_bmp = SelectObject(dc_bmp,g_display);
			if (cfg_toggle_scroll) {
				RECT r, r2;
				r.left = g_rect.right - g_column;
				r.right = g_rect.right;
				r.top = ps.rcPaint.top;
				r.bottom = ps.rcPaint.bottom;
				if (IntersectRect(&r2, &r, &ps.rcPaint)) {
					BitBlt(dc, r2.left, r2.top,
						r2.right - r2.left, r2.bottom - r2.top,
						dc_bmp,
						r2.left - r.left, ps.rcPaint.top, SRCCOPY);
				}
				r.left = 0;
				r.right = g_rect.right - g_column;
				r.top = ps.rcPaint.top;
				r.bottom = ps.rcPaint.bottom;
				if (IntersectRect(&r2, &r, &ps.rcPaint)) {
					BitBlt(dc, r2.left, r2.top,
						r2.right - r2.left, r2.bottom - r2.top,
						dc_bmp,
						g_column + r2.left, r2.top, SRCCOPY);
				}
			} else {
				BitBlt(dc,ps.rcPaint.left,ps.rcPaint.top,
					ps.rcPaint.right-ps.rcPaint.left,ps.rcPaint.bottom-ps.rcPaint.top,
					dc_bmp,
					ps.rcPaint.left,ps.rcPaint.top,SRCCOPY);
			}
			SelectObject(dc_bmp,old_bmp);
			DeleteDC(dc_bmp);

			// scale
			if (ps.rcPaint.right > g_rect.left) {
				HPEN pen = CreatePen(PS_SOLID,0,cfg_color_grid_line);
				COLORREF color_grid_line2 = RGB(
					(GetRValue(cfg_color_grid_line)+GetRValue(cfg_color_grid_bg))/2,
					(GetGValue(cfg_color_grid_line)+GetGValue(cfg_color_grid_bg))/2,
					(GetBValue(cfg_color_grid_line)+GetBValue(cfg_color_grid_bg))/2);
				HPEN pen2 = CreatePen(PS_SOLID,0,color_grid_line2);
				HBRUSH brush = CreateSolidBrush(cfg_color_grid_bg);
				HGDIOBJ old_brush = SelectObject(dc, brush);
				unsigned n, m = g_last_srate / 2000;
				unsigned ch;
				int x = g_rect.right;
				RECT rc_scale;
				rc_scale.left = x;
				rc_scale.top = 0;
				rc_scale.right = x + 5;
				rc_scale.bottom = g_rect.bottom;
				FillRect(dc,&rc_scale,brush);
				HGDIOBJ old_pen = SelectObject(dc,pen2);
				gradient * grad = gradient::get("RGB linear");
				if (grad) {
					grad->add_point(+1.5, cfg_color_grid_bg);
					grad->add_point(-0.5, cfg_color_grid_line);
					for (ch = 0; ch < g_last_nch; ch++) {
						int hi = g_grid[ch*g_last_spectrum_size];
						int lo = g_grid[(ch+1)*g_last_spectrum_size];
						int n;
						for (n = lo; n < hi; n++) {
							SetPixelV(dc, x, n, grad->get_color((double)(n-lo)/(double)(hi-lo)));
						}
					}
					grad->service_release();
				} else {
					MoveToEx(dc,x,0,NULL);
					LineTo(dc,x,g_rect.bottom);
				}
				SelectObject(dc,pen);
				for (ch = 0; ch < g_last_nch; ch++)
				{
					for (n = 1; n <= m; n++)
					{
						int idx = ch*g_last_spectrum_size+(int)(g_last_spectrum_size*n*2000/g_last_srate);
						int y = g_grid[idx];
						int dx = (n % 10) ? ((n % 5) ? 2 : 3) : 5;
						MoveToEx(dc,x,y,NULL);
						LineTo(dc,x+dx,y);
					}
				}
				SelectObject(dc,old_brush);
				DeleteObject(brush);
				SelectObject(dc,old_pen);
				DeleteObject(pen2);
				DeleteObject(pen);
			}

			EndPaint(wnd,&ps);
		}
		break;
	case WM_CONTEXTMENU:
		{
			enum {
				ID_SETTINGS=100,
				ID_CLOSE,
				ID_HIDE_CAPTION,
				ID_SCROLL,

				ID_DISPLAY_BASE = 1000,
				ID_RENDER_BASE = 2000,
			};

			HMENU menu = CreatePopupMenu();

			uAppendMenu(menu, 
				cfg_toggle_scroll ? MF_STRING|MF_CHECKED : MF_STRING,
				ID_SCROLL, "Enable scrolling");

			uAppendMenu(menu, MF_SEPARATOR, 0, 0);

			uAppendMenu(menu, 
				(cfg_display_mode == DISPLAY_ALL_CHANNELS) ? MF_STRING|MF_CHECKED : MF_STRING,
				ID_DISPLAY_BASE+DISPLAY_ALL_CHANNELS, "All channels");
			uAppendMenu(menu, 
				(cfg_display_mode == DISPLAY_MIXDOWN) ? MF_STRING|MF_CHECKED : MF_STRING,
				ID_DISPLAY_BASE+DISPLAY_MIXDOWN, "Mixdown");
			uAppendMenu(menu,
				(cfg_display_mode == DISPLAY_DIFFERENCE) ? MF_STRING|MF_CHECKED : MF_STRING,
				ID_DISPLAY_BASE+DISPLAY_DIFFERENCE, "Mean squared difference");

			uAppendMenu(menu, MF_SEPARATOR, 0, 0);

			uAppendMenu(menu, (cfg_interpolation == 0) ? MF_STRING|MF_CHECKED : MF_STRING,
				ID_RENDER_BASE+0, "No interpolation");

			uAppendMenu(menu, (cfg_interpolation == 1) ? MF_STRING|MF_CHECKED : MF_STRING,
				ID_RENDER_BASE+1, "Linear interpolation");

			uAppendMenu(menu, (cfg_interpolation == 2) ? MF_STRING|MF_CHECKED : MF_STRING,
				ID_RENDER_BASE+2, "Antialiasing");

			uAppendMenu(menu, MF_SEPARATOR, 0, 0);

			uAppendMenu(menu, MF_STRING, ID_SETTINGS, "Settings...");
			uAppendMenu(menu, MF_STRING, ID_HIDE_CAPTION, cfg_toggle_show_caption ? "Hide caption" : "Show caption");

			uAppendMenu(menu, MF_SEPARATOR, 0, 0);

			uAppendMenu(menu, MF_STRING, ID_CLOSE, "Close window");

			POINT p;
			GetCursorPos(&p);

			int cmd = TrackPopupMenu(menu, TPM_RIGHTBUTTON | TPM_NONOTIFY | TPM_RETURNCMD, p.x, p.y, 0, wnd, 0);

			DestroyMenu(menu);

			if (cmd == ID_SETTINGS)
				play_control::g_show_config("Simple spectrum");
			else if (cmd == ID_CLOSE)
				uPostMessage(wnd, WM_CLOSE, 0, 0);
			else if (cmd == ID_HIDE_CAPTION) {
				cfg_toggle_show_caption = !cfg_toggle_show_caption;
				DestroyWindow(wnd);
				create_vis_window();
			}
			else if (cmd == ID_SCROLL)
				cfg_toggle_scroll = !cfg_toggle_scroll;
			else if (cmd >= ID_RENDER_BASE)
				cfg_interpolation = cmd-ID_RENDER_BASE;
			else if (cmd >= ID_DISPLAY_BASE)
				cfg_display_mode = cmd-ID_DISPLAY_BASE;
			RedrawWindow(wnd, NULL, NULL, RDW_INVALIDATE);
		}
		break;
	case WM_CREATE:
		{
			g_wnd = wnd;
			dragging = false;
			apply_window_placement(cfg_window_placement, wnd);
			cfg_display_mode.refresh();
			if (cfg_toggle_use_alpha && g_SetLayeredWindowAttributes)
				g_SetLayeredWindowAttributes(wnd, 0, 255-cfg_alpha, LWA_ALPHA);
		}
		break;
	case WM_CLOSE:
		cfg_enabled = false;
		g_close();
		break;
	case WM_DESTROY:
		{
			read_window_placement(cfg_window_placement, wnd);
			g_wnd = 0;
		}
		break;
	case WM_KEYDOWN:
		keyboard_shortcut_manager::get()->on_keydown_auto(wp);
		break;
	case WM_SYSKEYDOWN:
		keyboard_shortcut_manager::get()->on_keydown_auto(wp);
		break;
	case WM_LBUTTONDOWN:
		{
			dragging = true;
			GetCursorPos(&drag_start);
			SetCapture(wnd);
		}
		break;
	case WM_LBUTTONUP:
		{
			ReleaseCapture();
		}
		break;
	case WM_CAPTURECHANGED:
		{
			dragging = false;
		}
		break;
	case WM_MOUSEMOVE:
		if (dragging)
		{
			POINT p;
			GetCursorPos(&p);
			RECT rc;
			GetWindowRect(wnd, &rc);
			POINT pos;
			pos.x = rc.left+p.x-drag_start.x;
			pos.y = rc.top+p.y-drag_start.y;
			SetWindowPos(wnd, 0, pos.x, pos.y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
			drag_start = p;
		}
		break;
	case WM_SIZE:
		GetClientRect(wnd, &g_rect);
		g_rect.right -= 5;
		break;
	}
	return uDefWindowProc(wnd,msg,wp,lp);
}


// updates precomputed gradient table
static void recalc_colors()
{
	gradient * cmap = gradient::get(cfg_gradient);
	if (cmap)
	{
		cmap->add_point(0.0, cfg_color_spectrum_low);
		cmap->add_point(1.0, cfg_color_spectrum_high);
		unsigned n;
		for (n=0;n<VIS_COLOR_TABLE_SIZE;n++)
		{
			g_colors[n] = cmap->get_color(n / ((double)VIS_COLOR_TABLE_MAX));
		}
		cmap->service_release();
	}
}

static int g_get_color_idx(double val) {
	/*
	val = log((val/2) + 1) / log(2);
	val *= 0.25;
	if (val>1.0) val = 1.0;
	int color_idx = (int)(val * VIS_COLOR_TABLE_MAX);
	color_idx = min(VIS_COLOR_TABLE_MAX,max(0,color_idx));
	*/

	static double log2_div = 1.0/log(2);
    val = sqrt(log((val * 0.5) + 1)) * log2_div;
    val *= 0.5;
    if (val > 1.0) val = 1.0;

    int color_idx = (int)(val * VIS_COLOR_TABLE_MAX);
    color_idx = min(VIS_COLOR_TABLE_MAX,max(0,color_idx));
    return color_idx;
}

class NOVTABLE slice_renderer {
public:
	virtual ~slice_renderer() {}

	virtual void init(HDC dc, int spectrum_size, int f_low, int f_high)=0;
	virtual void begin_slice(int y)=0;
	virtual void end_slice()=0;
	virtual void next_band(COLORREF c)=0;
};

class slice_renderer_simple : public slice_renderer {
private:
	HDC dc;
	int f_low, f_high;
	int spectrum_size;
	int y, dy;
	int x;
	double scale;
	double pos;
public:
	virtual void init(HDC dc, int spectrum_size, int f_low, int f_high) {
		this->dc = dc;
		this->f_low = f_low;
		this->f_high = f_high;
		this->spectrum_size = spectrum_size;
		this->scale = (abs(f_high - f_low) + 1) / (double)spectrum_size;
		this->dy = (f_low < f_high) ? 1 : -1;
	}
	virtual void begin_slice(int x) {
		this->x = x;
		y = f_low;
		pos = 0;
	}
	virtual void end_slice() {}
	virtual void next_band(COLORREF c) {
		double p = pos + scale;
		while (p >= 1) {
			SetPixelV(dc, x, y, c);
			y += dy;
			p -= 1;
		}
		pos = p;
	}
};

class slice_renderer_linear : public slice_renderer {
private:
	HDC dc;
	int f_low, f_high;
	int spectrum_size;
	int y, dy;
	int x;
	double r, g, b;
	double pos;
	double scale, inv_scale;
public:
	virtual void init(HDC dc, int spectrum_size, int f_low, int f_high) {
		this->dc = dc;
		this->f_low = f_low;
		this->f_high = f_high;
		this->spectrum_size = spectrum_size;
		this->scale = (abs(f_high - f_low) + 1) / (double)spectrum_size;
		this->inv_scale = 1.0 / this->scale;
		this->dy = (f_low < f_high) ? 1 : -1;
	}
	virtual void begin_slice(int x) {
		this->x = x;
		y = f_low;
		pos = -1;
	}
	virtual void end_slice() {
		if (pos > 0 && (y-f_high)*dy < 0) {
			COLORREF c2 = RGB(
				(int)r,
				(int)g,
				(int)b);
			SetPixelV(dc, x, y, c2);
		}
	}
	virtual void next_band(COLORREF c) {
		double r2, g2, b2;
		r2 = GetRValue(c);
		g2 = GetGValue(c);
		b2 = GetBValue(c);
		double p = pos + scale;
		if (pos < 0) {
			pos = 0;
			SetPixelV(dc, x, y, c);
			y += dy;
		} else if (p >= 1) {
			double f = 1 - pos;
			while (p >= 1) {
				COLORREF c2 = RGB(
					(int)(((scale-f)*r + r2*f)*inv_scale),
					(int)(((scale-f)*g + g2*f)*inv_scale),
					(int)(((scale-f)*b + b2*f)*inv_scale));
				SetPixelV(dc, x, y, c2);
				if (y == f_high)
					y = y;
				y += dy;
				p -= 1;
				f += 1;
			}
			pos = p;
		} else {
			pos += scale;
			r = r2;
			g = g2;
			b = b2;
		}
		r = r2;
		g = g2;
		b = b2;
	}
};

class slice_renderer_aa : public slice_renderer {
private:
	HDC dc;
	int f_low, f_high;
	int spectrum_size;
	int y, dy;
	int x;
	double r, g, b;
	double fill;
	double scale;
public:
	virtual void init(HDC dc, int spectrum_size, int f_low, int f_high) {
		this->dc = dc;
		this->f_low = f_low;
		this->f_high = f_high;
		this->spectrum_size = spectrum_size;
		this->scale = (abs(f_high - f_low) + 1) / (double)spectrum_size;
		this->dy = (f_low < f_high) ? 1 : -1;
	}
	virtual void begin_slice(int x) {
		this->x = x;
		y = f_low;
		fill = 0;
	}
	virtual void end_slice() {
		if (fill > 0 && (y-f_high)*dy < 0) {
			COLORREF c2 = RGB(
				(int)r,
				(int)g,
				(int)b);
			SetPixelV(dc, x, y, c2);
		}
	}
	virtual void next_band(COLORREF c) {
		double r2, g2, b2;
		r2 = GetRValue(c);
		g2 = GetGValue(c);
		b2 = GetBValue(c);
		double f = fill + scale;
		if (f >= 1) {
			if (fill > 0 && fill < 1) {
				COLORREF c2 = RGB(
					(int)(r + r2 * (1-fill)),
					(int)(g + g2 * (1-fill)),
					(int)(b + b2 * (1-fill)));
				SetPixelV(dc, x, y, c2);
				y += dy;
				f -= 1;
			}
			while (f >= 1) {
				SetPixelV(dc, x, y, c);
				y += dy;
				f -= 1;
			}
			fill = f;
			r = r2 * f;
			g = g2 * f;
			b = b2 * f;
		} else {
			fill = f;
			r += r2 * scale;
			g += g2 * scale;
			b += b2 * scale;
		}
	}
};

class visualization_simple_spectrum : public visualization
{
	double expected_latency;
	double timer_freq_div;
	int lagged_chunks;
public:
	visualization_simple_spectrum() {
		expected_latency = 0.0;
		LARGE_INTEGER timer_freq;
		if (QueryPerformanceFrequency(&timer_freq))
			timer_freq_div = 1.0 / (double)timer_freq.QuadPart;
		else
			timer_freq_div = 0.0;
		lagged_chunks = 0;
	}

	virtual bool is_active() {return !!cfg_enabled;}
	virtual bool need_spectrum() {return true;}
	virtual double get_expected_latency() {return 0 /*expected_latency*/ ;}


	virtual void on_data(const vis_chunk * p_data)
	{
		if (!cfg_enabled) return;
#if 0
		if (g_wnd && core_api::is_main_thread())
		{
			string8 caption;
			caption.add_string("lagged_chunks = ");
			caption.add_int(lagged_chunks);
			caption.add_string(", expected_latency = ");
			caption.add_float(expected_latency, 6);
			uSetWindowText(g_wnd,caption);
		}
#endif
		expected_latency /= 2;
		if (p_data->flags & vis_chunk::FLAG_LAGGED)
		{
			if (lagged_chunks++ >= cfg_lagged_chunks_threshold) return;
		}
		else
			lagged_chunks = 0;
		if (!g_wnd)
		{
			create_vis_window();
			if (!g_wnd) return;
			ShowWindow(g_wnd,SW_SHOWNOACTIVATE);
		}
		else if (g_timer)
		{
			KillTimer(0,g_timer);
			g_timer = 0;
		}

		int use_custom_fft = 0;
		vis_chunk * data = (vis_chunk *) p_data;
		if (use_custom_fft) {
			spectrum * sp = spectrum::create("Plain");
			if (sp) {
				data = new vis_chunk;
				sp->run(p_data, data);
				sp->service_release();
			}
		}

		if (g_display.resize(g_rect.right, g_rect.bottom))
		{
			recalc_grid(1,1); // hack
			recalc_colors();
			if (g_column >= (unsigned)g_rect.right)
				g_column = 0;
		}

		if (g_last_srate != data->srate)
		{
			g_last_srate = data->srate;
			HDC dc = CreateCompatibleDC(0);
			HGDIOBJ meh = SelectObject(dc,g_display);
			HBRUSH brush = (HBRUSH) GetStockObject(BLACK_BRUSH);
			FillRect(dc,&g_rect,brush);
			SelectObject(dc,meh);
			DeleteObject(brush);
			DeleteDC(dc);
			g_column = 0;
		}


		LARGE_INTEGER time_start,time_end;
		QueryPerformanceCounter(&time_start);

		{
			HDC dc = CreateCompatibleDC(0);
			HGDIOBJ meh = SelectObject(dc,g_display);

			RECT r;
			r.left = g_column;
			r.right = g_column+1;

			bool b_mono = false;
			bool b_mixdown = false;

			switch (cfg_display_mode)
			{
			case DISPLAY_MIXDOWN:
				b_mixdown = true;
				// fall through
			case DISPLAY_DIFFERENCE:
				b_mono = true;
				break;
			}

			slice_renderer * slice;
			switch (cfg_interpolation) {
			case 1:
				slice = new slice_renderer_linear();
				break;
			case 2:
				slice = new slice_renderer_aa();
				break;
			default:
				slice = new slice_renderer_simple();
			}

			if (b_mono)
			{
				recalc_grid(1, data->spectrum_size);
				int * rc = g_grid.get_ptr();
				unsigned n;
				vis_sample * src = data->spectrum;
				double div = 1.0 / (double) data->nch;
				mem_block_t<vis_sample> offset;
				offset.set_size(data->nch);
				for (n=0;n<data->nch;n++)
				{
					if (cfg_compensation)
						offset[n] = src[(data->spectrum_size-1)*data->nch+n];
					else
						offset[n] = 0.0;
				}
				slice->init(dc, data->spectrum_size, rc[0], rc[data->spectrum_size]);
				slice->begin_slice(r.left);
				for(n=0;n<data->spectrum_size;n++)
				{
					double val = 0;
					unsigned c;
					if (b_mixdown)
					{
						for(c=0;c<data->nch;c++)
						{
							val += fabs(*(src++)-offset[c]);
						}
					}
					else
					{
						double mean = 0;
						for(c=0;c<data->nch;c++)
						{
							mean += fabs(src[c]);
						}
						mean *= div;
						for(c=0;c<data->nch;c++)
						{
							double diff = mean-fabs(*(src++));
							val += diff*diff;
						}
					}
					val *= div;
					int color_idx = g_get_color_idx(val);
					slice->next_band(g_colors[color_idx]);
				}
				slice->end_slice();
			}
			else
			{
				recalc_grid(data->nch, data->spectrum_size);
				unsigned n;
				vis_sample * src = data->spectrum;
				mem_block_t<int *> rc;
				rc.set_size(data->nch);
				mem_block_t<vis_sample> offset;
				offset.set_size(data->nch);
				for (n=0;n<data->nch;n++)
				{
					// channels in g_grid are from bottom to top
					rc[n] = g_grid.get_ptr()+((data->nch-n-1)*data->spectrum_size);
					if (cfg_compensation)
						offset[n] = src[(data->spectrum_size-1)*data->nch+n];
					else
						offset[n] = 0.0;
				}

				unsigned c;
				for (c = 0; c < data->nch;  c++) {
					slice->init(dc, data->spectrum_size, rc[c][0], rc[c][data->spectrum_size]);
					slice->begin_slice(r.left);
					src = &data->spectrum[c];
					double val;
					for (n = 0; n < data->spectrum_size; n++) {
						val = fabs(*src);
						src += data->nch;
						val = fabs(val-offset[c]);
						int color_idx = g_get_color_idx(val);
						slice->next_band(g_colors[color_idx]);
					}
					slice->end_slice();
				}
			}

			delete slice;

			g_column = (g_column+1)%g_rect.right;

			SelectObject(dc,meh);
			DeleteDC(dc);

			r.top = 0;
			r.bottom = g_rect.bottom;

			RedrawWindow(g_wnd, cfg_toggle_scroll ? &g_rect : &r, NULL, RDW_INVALIDATE|RDW_UPDATENOW);
		}

		QueryPerformanceCounter(&time_end);
		expected_latency += (double)(time_end.QuadPart-time_start.QuadPart)*timer_freq_div/2;

		if (use_custom_fft) {
			delete data->spectrum;
			delete data;
		}


	}
	virtual void on_flush()
	{
		lagged_chunks = 0;
		g_flush();
	}
};

class menu_item_simple_spectrum : public menu_item_main
{
	virtual unsigned get_num_items() {
		return 1;
	}

	virtual void enum_item(unsigned n, string_base & out) {
		if (n == 0) {
			out = "Components/Visualization/Simple spectrum";
		}
	}

	virtual void perform_command(unsigned n) 
	{
		if (n==0 && core_api::assert_main_thread())
		{
			cfg_enabled = !cfg_enabled;
			if (!cfg_enabled) g_close();
			else
			{
				if (!visualization::is_vis_manager_present()) console::error("Visualisation manager not present, please reinstall.");
			}
		}
	}
	virtual bool is_checked(int n) {return n==0 ? !!cfg_enabled : 0;};
};

static bool choose_color_help(HWND parent, cfg_int & cfg_color) {
	DWORD custom_colors[16] = {cfg_color_spectrum_low, cfg_color_spectrum_high, cfg_color_grid_line, cfg_color_grid_bg};
	DWORD color = cfg_color;
	if (uChooseColor(&color,parent,custom_colors))
	{
		cfg_color = color;
		return true;
	}
	return false;
}

class config_simple_spectrum : public config
{
	static BOOL CALLBACK ConfigProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
	{
		static ptr_list_simple<string_simple> list;

		switch(msg)
		{
		case WM_INITDIALOG:
			{
				list.delete_all();
				unsigned n = 0;
				bool found = false;
				service_enum_t<gradient> e;
				gradient * ptr;
				for (ptr=e.first();ptr;ptr=e.next())
				{
					uSendDlgItemMessageText(wnd,IDC_GRADIENT,CB_ADDSTRING,0,ptr->get_name());
					list.add_item(new string_simple(ptr->get_name()));
					if (!strcmp(cfg_gradient,ptr->get_name()))
					{
						uSendDlgItemMessage(wnd,IDC_GRADIENT,CB_SETCURSEL,n,0);
						found = true;
					}
					ptr->service_release();
					n++;
				}
				if (!found)
				{
					uSendDlgItemMessage(wnd,IDC_GRADIENT,CB_SETCURSEL,0,0);
					cfg_gradient = list[0]->get_ptr();
					recalc_colors();
				}
				uSendDlgItemMessage(wnd,IDC_SLIDER1,TBM_SETRANGEMIN,0,0);
				uSendDlgItemMessage(wnd,IDC_SLIDER1,TBM_SETRANGEMAX,0,500);
				uSendDlgItemMessage(wnd,IDC_SLIDER1,TBM_SETPOS,1,cfg_range);
				if (cfg_range)
				{
					uSendDlgItemMessageText(wnd,IDC_SCALEINFO,WM_SETTEXT,0,
						string_printf("logarithmic scale, x = %d.%02d",cfg_range/100,cfg_range%100));
				}
				else
					uSendDlgItemMessageText(wnd,IDC_SCALEINFO,WM_SETTEXT,0,"linear scale");
				console::info(string_printf("setting %d",(int)cfg_lagged_chunks_threshold));
				SetDlgItemInt(wnd,IDC_LAGGEDCHUNKS,(int)cfg_lagged_chunks_threshold,1);
				uSendDlgItemMessage(wnd,IDC_WINDOWING,BM_SETCHECK,cfg_compensation,0);

				if (g_SetLayeredWindowAttributes) {
					EnableWindow(GetDlgItem(wnd, IDC_ENABLETRANSPARENCY), true);
				}
				uSendDlgItemMessage(wnd, IDC_ENABLETRANSPARENCY, BM_SETCHECK, cfg_toggle_use_alpha ? BST_CHECKED : BST_UNCHECKED, 0);
				uSendDlgItemMessage(wnd, IDC_ALPHA, TBM_SETRANGEMIN, 0, 0);
				uSendDlgItemMessage(wnd, IDC_ALPHA, TBM_SETRANGEMAX, 0, 255);
				uSendDlgItemMessage(wnd, IDC_ALPHA, TBM_SETPOS, 1, cfg_alpha);

			}
			break;
		case WM_HSCROLL:
			switch (LOWORD(wp))
			{
			case TB_THUMBTRACK:
			case TB_ENDTRACK:
				if (((HWND)lp) == GetDlgItem(wnd, IDC_ALPHA)) {
					cfg_alpha = uSendDlgItemMessage(wnd, IDC_ALPHA, TBM_GETPOS, 0, 0);
					if (cfg_toggle_use_alpha && g_wnd && g_SetLayeredWindowAttributes)
						g_SetLayeredWindowAttributes(g_wnd, 0, 255-cfg_alpha, LWA_ALPHA);
				} else{
					cfg_range = uSendDlgItemMessage(wnd,IDC_SLIDER1,TBM_GETPOS,0,0);
					if (cfg_range)
					{
						uSendDlgItemMessageText(wnd,IDC_SCALEINFO,WM_SETTEXT,0,
							string_printf("logarithmic scale, x = %d.%02d",cfg_range/100,cfg_range%100));
					}
					else
						uSendDlgItemMessageText(wnd,IDC_SCALEINFO,WM_SETTEXT,0,"linear scale");
				}
				break;
			}
			break;
		case WM_COMMAND:
			switch(wp)
			{
			case IDC_WINDOWING | (BN_CLICKED<<16):
				cfg_compensation = uSendMessage((HWND)lp,BM_GETCHECK, 0, 0);
				break;
			case IDC_ENABLETRANSPARENCY | (BN_CLICKED<<16):
				cfg_toggle_use_alpha = uSendMessage((HWND)lp, BM_GETCHECK, 0, 0);
				if (cfg_toggle_use_alpha && g_wnd && g_SetLayeredWindowAttributes) {
					uSetWindowLong(g_wnd, GWL_EXSTYLE,
						uGetWindowLong(g_wnd, GWL_EXSTYLE) | WS_EX_LAYERED);
					g_SetLayeredWindowAttributes(g_wnd, 0, 255-cfg_alpha, LWA_ALPHA);
				} else {
					uSetWindowLong(g_wnd, GWL_EXSTYLE,
						uGetWindowLong(g_wnd, GWL_EXSTYLE) & ~WS_EX_LAYERED);
				}
				break;
			case IDC_LAGGEDCHUNKS | (EN_CHANGE<<16):
				{
					int bleh;
					cfg_lagged_chunks_threshold = (int)GetDlgItemInt(wnd,IDC_LAGGEDCHUNKS,&bleh,1);
					console::info(string_printf("read %d",(int)cfg_lagged_chunks_threshold));
				}
				break;
			case IDC_COLOR_LOW:
				choose_color_help(wnd, cfg_color_spectrum_low);
				break;
			case IDC_COLOR_HIGH:
				choose_color_help(wnd, cfg_color_spectrum_high);
				break;
			case IDC_COLOR_GRID_LINE:
				choose_color_help(wnd, cfg_color_grid_line);
				break;
			case IDC_COLOR_GRID_BG:
				choose_color_help(wnd, cfg_color_grid_bg);
				break;
			case IDC_GRADIENT | (CBN_SELCHANGE<<16):
				{
					unsigned n = uSendDlgItemMessage(wnd,IDC_INTERPOLATION,CB_GETCURSEL,0,0);
					cfg_gradient = list[n]->get_ptr();
					recalc_colors();
				}
				break;
			}
			break;
		}
		return 0;
	}
public:
	virtual HWND create(HWND parent)
	{
		return uCreateDialog(IDD_CONFIG,parent,ConfigProc);
	}
	virtual const char * get_name() {return "Simple spectrum";}
	virtual const char * get_parent_name() {return "Visualizations";}
};

class initquit_simple_spectrum : public initquit
{
public:
	virtual void on_init() {
		init_layered_window();
	}
	virtual void on_quit()
	{
		if (g_wnd) DestroyWindow(g_wnd);
	}
};

static visualization_factory<visualization_simple_spectrum> foo;
static menu_item_factory<menu_item_simple_spectrum> foo2;
static config_factory<config_simple_spectrum> foo3;
static initquit_factory<initquit_simple_spectrum> foo4;

static gradient_factory<gradient_rgb_linear> foo5;
static gradient_factory<gradient_hsv_ccw> foo6;
static gradient_factory<gradient_hsv_cw> foo7;
